(* (C) ean-Franois Monin, 1999                          *)
(* France Telecom R & D                                  *)
(* $Id: catags.ml,v 1.1 2001/09/26 16:57:06 alvaracu Exp $ *)

let help_string = "
Usage:
$ catags file1 file2 ...

where file1, file2, ... are tag files analysed as sequences of lines
-----------------
^L
path_x,nbx
line1
line2
...
lineX
^L
path_y,nby
line1
line2
...
lineY

(and so on.)
-----------------

outputs on stdout the catenation of file1 file2 ...
where paths like path_x which are relative
become relative to the current directory.
For instance, if file1 is xyz/TAGS and contains
foo.c,459
and
../abc/bar.c,843

the corresponding lines will be replaced with

xyz/foo.c,459
and
abc/bar.c,843
"

let err s =
  let prog = Sys.argv.(0) in
  Printf.fprintf stderr "Error in %s;\n%s\n%s\n"
    prog "complain to JeanFrancois.Monin@francetelecom.com" s;
  flush stderr

let split s =
  let fromto i j = String.sub s i (j-i) in
  let rec tronconne n =
    try
      let k = String.index_from s n '/' + 1 in fromto n k :: tronconne k
    with Not_found -> [fromto n (String.length s)]
  in tronconne 0

let recolle = List.fold_left (^) ""

(* Elimination of subsequences "../" "xxx" and of "./"            *)
(* normalise2 r1 l = r2  implies  normalise (rev r1 ^ l) = rev r2 *)
let rec normalise2 r1 l2 =
  match r1, l2 with
  | r1, "./"::l2 -> normalise2 r1 l2
  | x1::r1, "../"::l2 when x1 <> "../" -> normalise2 r1 l2
  | r1, x2::l2 -> normalise2 (x2::r1) l2
  | r1, [] -> r1

let conc_chem revdirlist path = 
  let l = split path in
  if List.hd l = "/" then path else recolle (List.rev (normalise2 revdirlist l))

let revdirlist path = List.tl (normalise2 [] (split path))

let cut_path_nb l =
  let i = String.rindex l ',' in
  String.sub l 0 i, String.sub l i (String.length l - i)

let transf f cin cout =
  try while true do 
    let l = input_line cin in
    Printf.fprintf cout "%s\n" l;   
    if l = "" then
      let path,nb = cut_path_nb (input_line cin) in
      Printf.fprintf cout "%s%s\n" (f path) nb
  done
  with End_of_file -> flush cout

exception Help

let list_args () =
  let rec list_args args n =
    if n = Array.length args then []
    else let s = args.(n) in
    if s = "-help" or s = "--help" or s = "-h" then raise Help
    else s::list_args args (n+1)
  in list_args Sys.argv 1

(*
let args_iter f =
  let args = Sys.argv in 
  for i = 1 to Array.length args - 1 do f args.(i) done
*)

let main =
  let to_be_applied s = 
    let cin = open_in s in
    transf (conc_chem (revdirlist s)) cin stdout; close_in cin in
  try List.iter to_be_applied (list_args ())
  with 
  | Help -> prerr_endline help_string 
  | x -> err ""; raise x
