(* (C) Cuihtlauac Alvarado, Jean-Franois Monin  1999-2001  *)
(*     France Telecon Recherche & Developpement             *)

module type Tags_sig = sig
  val add : string -> int * int -> unit
end

module Tags = functor (T : Tags_sig) -> struct

let log s (x,y) =
  Printf.fprintf stderr "Warning : %s at %d-%d\n" s x y; flush stderr

let rec pe_patt ast = 
  let loc = MLast.loc_of_patt ast in
  match ast with
  | <:patt< ( $p1$ as $p2$ ) >> -> pe_patt p1; pe_patt p2 
  | <:patt< $lid:i$ >> -> T.add i loc
  | <:patt< ( $list:pl$ ) >> -> List.iter pe_patt pl
  | <:patt< ( $p$ : $t$ ) >> -> pe_patt p
  | <:patt< $_$ >> -> ()
(*
  | _ -> log "patt forgotten" loc; ()
*)

let rec pe_ctyp ast = 
  let loc = MLast.loc_of_ctyp ast in
  match ast with
  | <:ctyp< $t1$ as $t2$ >> -> pe_ctyp t1; pe_ctyp t2
  | <:ctyp< $t1$ == $t2$ >> -> pe_ctyp t1; pe_ctyp t2
  | <:ctyp< { $list:sbtl$ } >> ->
      List.iter (fun (loc,s,_,c) -> T.add s loc; pe_ctyp c) sbtl
  | <:ctyp< [ $list:stll$ ] >> ->
      List.iter (fun (loc,s,l) -> T.add s loc; List.iter pe_ctyp l) stll
  | _ -> ()


let pe_class_str_item ast = 
  let loc = MLast.loc_of_class_str_item ast in 
  match ast with
  | <:class_str_item< value $mut:mf$ $lab$ = $e$ >> -> T.add lab loc
  | <:class_str_item< method virtual private $s$ : $t$ >> -> T.add s loc
  | <:class_str_item< method virtual $l$ : $t$ >> -> T.add l loc
  | <:class_str_item< method private $l$ = $fb$ >> -> T.add l loc
  | <:class_str_item< method $l$ = $fb$ >> -> T.add l loc
  | _ -> ()


let rec pe_class_expr ast =
  let loc = MLast.loc_of_class_expr ast in
  match ast with
  | <:class_expr< object $cspo$ $list:cf$ end >> ->
      List.iter pe_class_str_item cf
  | <:class_expr< let $rec:r$ $list:_$ in $ce$ >> -> pe_class_expr ce
  | <:class_expr< $ce$ $expr$ >> -> pe_class_expr ce
  | <:class_expr< fun $p$ -> $cfb$ >> -> pe_class_expr cfb
  | <:class_expr< ($ce$ : $ct$) >> -> pe_class_expr ce
  | <:class_expr< $list:id$ [ $list:tl$ ] >> -> ()
(* next one is possible but not neccessary
 *  | <:class_expr< $list:id$ >> -> ()
 *)

let etag_class_info {MLast.ciNam = s; MLast.ciExp = c; MLast.ciLoc = loc} =
  T.add s loc; pe_class_expr c


let rec pe_class_sig_item ast = 
  let loc = MLast.loc_of_class_sig_item ast in
  match ast with
					(* CgCtr *)
    | <:class_sig_item< type $t1$ = $t2$ >> -> 
	()
    | <:class_sig_item< declare $list:csil$ end >> ->
	List.iter pe_class_sig_item csil
					(* CgInh *)
    | <:class_sig_item< inherit $cs$ >> -> 
	()
					(* CgMth *)
    | <:class_sig_item< method private $name$ : $t$ >> -> 
	T.add name loc
    | <:class_sig_item< method $name$ : $t$ >> -> 
	T.add name loc
					(* CgVal *)
    | <:class_sig_item< value $mut:mf$ $name$ : $t$ >> -> 
	T.add name loc
					(* CgVir *)
    | <:class_sig_item< method virtual private $name$ : $t$ >> -> 
	T.add name loc
    | <:class_sig_item< method virtual $name$ : $t$ >> -> 
	T.add name loc


let rec pe_class_type ast =
  let loc = MLast.loc_of_class_type ast in
  match ast with
    | <:class_type< $list:id$ [ $list:tl$ ] >> -> ()
(* possible but not necessary
 *     | <:class_type< $list:id$ >> -> ()
 *)

(* the following cannot occur here: functional class types are only
 * allowed in type annotations, not in class type definitions 
 *)
    | <:class_type< [ $typ$ ] -> $ct$ >> -> ()

    | <:class_type< object $opt:cst$ $list:csf$ end >> -> 
	List.iter pe_class_sig_item csf


let etag_class_type_info   
  {MLast.ciNam = s; MLast.ciExp = c; MLast.ciLoc = loc} =
  T.add s loc; pe_class_type c

let rec pe_str_item ast = 
  let loc = MLast.loc_of_str_item ast in
  match ast with
  | <:str_item< declare $list:stl$ end >> -> List.iter pe_str_item stl
  | <:str_item< exception $s$ of $list:tl$ >> -> T.add s loc
  | <:str_item< external $s$ : $t$ = $list:sl$ >> -> T.add s loc
  | <:str_item< module $s$ = $me$ >> -> T.add s loc; pe_module_expr me
  | <:str_item< module type $i$ = $mt$ >> -> T.add i loc
  | <:str_item< type $list:ssltl$ >>  ->
      List.iter (fun ((loc,s),_,c,_) -> T.add s loc; pe_ctyp c) ssltl
  | <:str_item< value $rec:b$ $list:pel$ >> ->
	  List.iter (fun (p,  _) -> pe_patt p) pel
  | <:str_item< class $list:cd$ >> ->
      List.iter (etag_class_info) cd
  | <:str_item< class type $list:ctd$ >> ->
      List.iter (etag_class_type_info) ctd
  | _ -> ()

and pe_module_expr ast = 
  let loc = MLast.loc_of_module_expr ast in
  match ast with
  | <:module_expr< functor ( $i$ : $t$ ) -> $me$ >> -> pe_module_expr me
  | <:module_expr< struct $list:st$ end >> -> List.iter pe_str_item st
  | _ -> ()

and pe_sig_item ast = 
  let loc = MLast.loc_of_sig_item ast in
  match ast with
  | <:sig_item< declare $list:stl$ end >> -> List.iter pe_sig_item stl
  | <:sig_item< exception $s$ of $list:tl$ >> -> T.add s loc
  | <:sig_item< external $s$ : $t$ = $list:sl$ >> -> T.add s loc
  | <:sig_item< module $s$ : $mt$ >> -> T.add s loc
  | <:sig_item< module type $i$ = $mt$ >> -> T.add i loc
  | <:sig_item< type $list:ssltl$ >>  ->
      List.iter (fun ((loc,s),_,c,_) -> T.add s loc; pe_ctyp c) ssltl
  | <:sig_item< value $s$ : $t$ >> -> T.add s loc 
  | <:sig_item< class $list:cd$ >> ->
      List.iter (etag_class_type_info) cd
  | _ -> ()

let rec implem = function
  | [] -> ()
  | (a,_)::l -> pe_str_item a; implem l

let rec interf = function
  | [] -> ()
  | (a,_)::l -> pe_sig_item a; interf l

let _ = Pcaml.print_implem := implem
let _ = Pcaml.print_interf := interf

end

(* $Id: pr.ml,v 1.5 2002/10/11 08:44:37 alvaracu Exp $ *)

