(* (C) Cuihtlauac Alvarado, Jean-Franois Monin   1999 *)
(*     Centre National d'tudes des Tlcommunications *)

(* $Id: vi.ml,v 1.6 2002/11/13 15:40:59 alvaracu Exp $ *)

let log s n = Printf.fprintf stderr "pb etag : %s, %d\n" s n; flush stderr

let sized_string_of_file f taille =
  let buf = String.create taille in
  let cin = open_in f in
  let rec loop lus0 t0 =
    let n = input cin buf lus0 t0 in
    let lus = lus0 + n in let t = t0 - n in 
    if n > 0 && t > 0 then loop lus t else lus in
  let lus = loop 0 taille in close_in cin;
  buf,lus

let make_strings f lg_file =
  let lg_with_nl = lg_file+1 in
  let fs,l = sized_string_of_file f lg_with_nl in
  fs.[l-1] <- '\n';
  (fs,l)

  
module Line : sig
  val lookup : string * int -> unit
  val of_pos : int -> int * int
  val print : unit -> unit
end = struct
  module LN = Line_number
  let pos_set = LN.create ()
  let lookup (file,len) =
    let add n = LN.add pos_set n in
    let rec look_nl n0 = 
      if n0 < len then begin
	let n = String.index_from file n0 '\n' in
	add n; look_nl (n+1)
      end else () in
    try look_nl 0 with Not_found -> ()
  let of_pos p =
    let x = LN.find pos_set p in
    LN.first x, LN.num x
  let print () = LN.print pos_set
end

module Entry = struct

  (* Store entries before writing them down. *)
  let lifo : (string * int * int) list ref = ref []

  let set_size_ml,size_ml =
    let c = ref 0 in
    let set n = c := n in
    let get () = !c
    in set,get

  let add etag (loc, last) =
    lifo := (etag, loc, last) :: !lifo; set_size_ml last

  let escape s =
    let part_es = String.escaped s in
    let buffer = Buffer.create (String.length part_es) in
    String.iter (fun c ->
                   match c with
                   | '$' | '/' | '\\' -> 
                       Buffer.add_char buffer '\\';
                       Buffer.add_char buffer c
                   | _ -> Buffer.add_char buffer c)
                part_es;
    Buffer.contents buffer        

  let line filename tagname linebeg =
    let etag =
      Printf.sprintf "%s\t%s\t/^%s$/;\n" tagname filename (escape linebeg) in
    etag, (String.length etag)

  let format filename (fs, l) =
    let rec loop lcur scur = function
      |	[] -> lcur,scur
      |	(entry, pos, last) :: rest ->
          let line_beg, _ = Line.of_pos pos in
          let len = 
            (try String.index_from fs line_beg '\n' 
             with Not_found -> String.length fs) - line_beg in
          let fl, n = 
             line filename entry (String.sub fs line_beg len) in
             loop (fl :: lcur) (n + scur) rest in 
    loop [] 0 !lifo

end

let add = Entry.add

let header chan = Printf.fprintf chan 
"!_TAG_FILE_FORMAT\t1\t/without ;\"/
!_TAG_FILE_SORTED\t0\t/0=unsorted, 1=sorted/
"

let process_file () =
  let arg = !Pcaml.input_file in
  let in_file = arg in
  let out_file = Sys.getenv Argcamlp4.tmp in 
  let name = Filename.basename in_file in
  let modulename = String.capitalize (Filename.chop_extension name) in
  let lg_file = Entry.size_ml () in
  let fsl = make_strings in_file lg_file in
  Line.lookup fsl;
  let lef, nef = Entry.format in_file fsl in
  let total = nef in
  let chan = open_out out_file in
  begin
    header chan;
    List.iter (output_string chan) lef;
    flush chan;
    close_out chan
  end

let _ = at_exit process_file

